@@ -13,9 +13,12 @@ from __future__ import unicode_literals, absolute_import, division, \
     print_function
 import os
 import functools
+import hashlib
 from flask import current_app, url_for
+from .exceptions import CompressorException
 from .blueprint import blueprint as compressor_blueprint
 from .templating import compressor as compressor_template_helper
+from .processors import DEFAULT_PROCESSORS
 
 
 class memoized(object):
@@ -57,12 +60,6 @@ class memoized(object):
         return functools.partial(self.__call__, obj)
 
 
-class CompressorException(Exception):
-    """ Base exception for all exceptions raised by the Flask-Compressor
-    extension. """
-    pass
-
-
 class Compressor(object):
     """
         The Flask extension object used by Flask-Compressor.
@@ -147,7 +144,6 @@ class Compressor(object):
         listed in `flask_compressor.processors.DEFAULT_PROCESSORS`.
         """
 
-        from .processors import DEFAULT_PROCESSORS
         for processor in DEFAULT_PROCESSORS:
             self.register_processor(processor)
 
@@ -214,9 +210,11 @@ class Bundle(object):
     default_linked_template = '<link ref="external" href="{url}" '\
                               'type="{mimetype}">'
     default_mimetype = 'text/plain'
+    default_extension = 'txt'
 
     def __init__(self, name, assets=None, processors=None,
-                 inline_template=None, linked_template=None, mimetype=None):
+                 inline_template=None, linked_template=None, mimetype=None,
+                 extension=None):
         """ Initializes a :class:`Bundle` instance.
 
         Args:
@@ -236,6 +234,8 @@ class Bundle(object):
                 `<link ref='external' href='{url}' type='{mimetype}'>`)
             mimetype: the mimetype corresponding to the final content of the
                 bundle (default: `text/plain`)
+            extension: the file extension associated with the mimetype
+                (default: `txt`)
         """
         self.name = name
         self.assets = assets or []
@@ -243,6 +243,10 @@ class Bundle(object):
         self.inline_template = inline_template or self.default_inline_template
         self.linked_template = linked_template or self.default_linked_template
         self.mimetype = mimetype or self.default_mimetype
+        self.extension = extension or self.default_extension
+
+        for asset in self.assets:
+            asset.bundle = self
 
     def apply_processors(self, contents):
         """ Apply all processors to the provided data.
@@ -339,21 +343,30 @@ class Bundle(object):
                     content. (default: `True`)
         """
         if concatenate:
-            url = url_for('compressor.render_bundle', bundle_name=self.name)
-            return self.linked_template.format(url=url, mimetype=self.mimetype)
+            return self.linked_template.format(url=self.url, mimetype=self.mimetype)
         else:
-            urls = []
-            for index, asset in enumerate(self.assets):
-                urls.append(url_for('compressor.render_asset',
-                                    bundle_name=self.name,
-                                    asset_index=index,
-                                    asset_name=asset.name))
-
+            urls = [asset.url for asset in self.assets]
             return '\n'.join(
                 [self.linked_template.format(url=url, mimetype=self.mimetype)
                  for url in urls]
             )
 
+    @property
+    @memoized
+    def url(self):
+        return url_for(
+            'compressor.render_bundle',
+            bundle_name=self.name,
+            bundle_hash=self.hash,
+            bundle_extension=self.extension,
+        )
+
+    @property
+    @memoized
+    def hash(self):
+        content = self.get_content()
+        return hashlib.md5(content.encode('utf-8')).hexdigest()
+
 
 class CSSBundle(Bundle):
     """ A helper class to use a :class:`Bundle` objects with CSS assets. """
@@ -361,6 +374,7 @@ class CSSBundle(Bundle):
     default_linked_template = '<link type="{mimetype}" rel="stylesheet" ' \
                               'href="{url}">'
     default_mimetype = 'text/css'
+    default_extension = 'css'
 
 
 class JSBundle(Bundle):
@@ -370,6 +384,7 @@ class JSBundle(Bundle):
     default_linked_template = '<script type="{mimetype}" src="{url}">' \
                               '</script>'
     default_mimetype = 'text/javascript'
+    default_extension = 'js'
 
 
 class Asset(object):
@@ -391,6 +406,7 @@ class Asset(object):
         """
         self._raw_content = content
         self.processors = processors or []
+        self.bundle = None
 
     def apply_processors(self, content):
         """ Apply all processors to the provided content.
@@ -426,6 +442,23 @@ class Asset(object):
         """ Return the raw content of the asset, without any modification. """
         return self._raw_content
 
+    @property
+    @memoized
+    def url(self):
+        return url_for(
+            'compressor.render_asset',
+            bundle_name=self.bundle.name,
+            bundle_extension=self.bundle.extension,
+            asset_index=self.bundle.assets.index(self),
+            asset_hash=self.hash,
+            name=self.name
+        )
+
+    @property
+    @memoized
+    def hash(self):
+        return hashlib.md5(self.content.encode('utf-8')).hexdigest()
+
 
 class FileAsset(Asset):
     """
